/**
 * Copyright (c) 2007-2012 EBM WebSourcing, 2012-2015 Linagora
 *
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.mail;

import java.io.UnsupportedEncodingException;
import java.util.Date;
import java.util.Set;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.NormalizedMessage;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Session;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.mail.internet.MimeUtility;

import org.ow2.petals.bc.mail.service.provide.ProvideDescriptor;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.w3c.dom.Node;

/**
 * This class is used to process (reading or creating) mail messages.
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 */
public class MimeMessageManager {

    protected Logger log;

    /**
     *
     * @param log
     *            the component {@link Logger}
     * @param channel
     *            the component {@link DeliveryChannel}
     */
    public MimeMessageManager(Logger log) {
        this.log = log;
    }

    /**
     * Create a {@link MimeMessage} from the given {@link NormalizedMessage},
     * including mail {@link Session} information.
     *
     * @param session
     *            the mail {@link Session}
     * @param provideDescriptor
     *            the {@link SessionDescriptor}
     * @param exchange
     *            the mapped {@link NormalizedMessage}
     * @param rootNode
     *            TODO
     * @return the created {@link MimeMessage}
     * @throws MessagingException
     *             if an error occurs during {@link MimeMessage} creation
     * @throws MissingElementException
     */
    public MimeMessage mapNormalizedMessageToMimeMessage(
            Session session,
            ProvideDescriptor provideDescriptor,
            Exchange exchange,
            Node rootNode )
    throws javax.jbi.messaging.MessagingException, MissingElementException {

        this.log.fine("create email from JBI Exchange");
        final MimeMessage mimeMessage = new MimeMessage(session);
        final String sendMode = provideDescriptor.getSendMode();

        // Set MimeMessage attributes
        this.setMimeMessageAttributes(
                mimeMessage,
                provideDescriptor.getToAddress(),
                provideDescriptor.getFromAddress(),
 provideDescriptor.getReplyAddress(),
                provideDescriptor.getSubject());

        try {
            // Extract and encode the content
            String content = MailUtil.extractMailBodyFromPayload(rootNode);
            try {
                content = MimeUtility.encodeText(content);
            } catch (UnsupportedEncodingException e) {
                throw new javax.jbi.messaging.MessagingException(e);
            }

            // No attachment to set
            if( exchange.getInMessageAttachmentNames().isEmpty()
                    || MailConstants.SEND_MODE_CONTENTONLY.equalsIgnoreCase( sendMode )) {
                mimeMessage.setContent( content, provideDescriptor.getContentType());
            }

            // There are attachments, we use multi-parts
            else {
                final Multipart multipart = new MimeMultipart();
                if( ! sendMode.equalsIgnoreCase(MailConstants.SEND_MODE_ATTACHMENTSONLY)) {
                    try {
                        MimeBodyPart messageBodyPart = new MimeBodyPart();
                        messageBodyPart.setContent( content, provideDescriptor.getContentType());
                        multipart.addBodyPart( messageBodyPart );

                    } catch( MessagingException e ) {
                        throw new javax.jbi.messaging.MessagingException(e);
                    }
                }

                this.setMimeMessageAttachments( multipart, exchange, rootNode );
                mimeMessage.setContent(multipart);
            }
        } catch (final javax.mail.MessagingException e) {
            final String msg = "Error setting Mime message content"
                    + "with all information collected from NormalizedMessage";
            throw new javax.jbi.messaging.MessagingException(msg, e);
        }

        return mimeMessage;
    }

    /**
     * Add {@link NormalizedMessage} attachments to {@link MimeMessage}
     * attachments. If an error occurs during the transmission of an attachment,
     * it is skipped and the process continues with other attachments.
     *
     * @param multipart
     *            the {@link Multipart} used to handle attachments
     * @param in
     *            the {@link NormalizedMessage} handling attachments that have
     *            to be transmitted
     */
    protected void setMimeMessageAttachments(Multipart multipart, Exchange exchange, Node rootNode)
            throws javax.jbi.messaging.MessagingException {
        this.log.fine("set the email attachments");
        NormalizedMessage in = exchange.getInMessage();
        final Set<String> attachmentIds = in.getAttachmentNames();
        MimeBodyPart attachementBodyPart = null;
        DataHandler handler = null;
        if (attachmentIds != null) {
            for (final String id : attachmentIds) {
                handler = in.getAttachment(id);
                this.log.fine("add " + id + " as MimeMessage attachment.");
                attachementBodyPart = new MimeBodyPart();
                try {
                    attachementBodyPart.setDataHandler(handler);
                    attachementBodyPart.setFileName(id);
                    multipart.addBodyPart(attachementBodyPart);
                } catch (final javax.mail.MessagingException e) {
                    final String msg = "Error setting MimeMessage attachments " + id;
                    this.log.warning(msg);
                }
            }
        }
    }

    /**
     * Sets {@link MimeMessage} recipient, date, from and subject.
     *
     * @param mimeMessage
     *            the {@link MimeMessage} to set
     * @param toAddress
     *            the destination e-mail address
     * @param fromAddress
     *            the sender e-mail address
     * @param replyAddress
     *            the reply e-mail address
     * @param subject
     *            the mail subject
     */
    protected void setMimeMessageAttributes(final MimeMessage mimeMessage, final String toAddress,
            final String fromAddress, final String replyAddress, final String subject)
            throws javax.jbi.messaging.MessagingException {
        this.log.fine("set the email attributes");
        try {
            // Build the destination address
            final InternetAddress toInternetAddress = new InternetAddress(toAddress);
            // Build the sender address
            final InternetAddress fromInternetAddress = new InternetAddress(fromAddress);
            mimeMessage.setRecipient(Message.RecipientType.TO, toInternetAddress);
            mimeMessage.setSentDate(new Date());
            mimeMessage.setFrom(fromInternetAddress);
            mimeMessage.setSubject(subject);
            if (replyAddress != null) {
                // Build the reply address
                final InternetAddress[] replyInternetAddresses = new InternetAddress[] { new InternetAddress(
                        replyAddress) };
                mimeMessage.setReplyTo(replyInternetAddresses);
            }

        } catch (final javax.mail.MessagingException e) {
            final String msg = "Error setting MimeMessage attributes";
            throw new javax.jbi.messaging.MessagingException(msg, e);
        }
    }
}
