/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2015 Linagora
 *
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.mail.listeners;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.Timer;

import javax.mail.Address;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Part;
import javax.mail.Session;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeUtility;
import javax.mail.util.SharedByteArrayInputStream;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.bc.mail.MailComponent;
import org.ow2.petals.bc.mail.MailConstants;
import org.ow2.petals.bc.mail.MailSessionManager;
import org.ow2.petals.bc.mail.MailUtil;
import org.ow2.petals.bc.mail.MimeMessageManager;
import org.ow2.petals.bc.mail.MissingElementException;
import org.ow2.petals.bc.mail.service.consume.ConsumeDescriptor;
import org.ow2.petals.bc.mail.service.consume.ExternalEventPoller;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractExternalListener;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;

/**
 * This class is used to manage listeners for external mail addresses registered during SU deployment.
 * @author Adrien LOUIS - EBM WebSourcing
 * @author Olivier Fabre - EBM WebSourcing
 */
public class ExternalListener extends AbstractExternalListener {

    protected ConfigurationExtensions extensions;

    protected MailSessionManager mailSessionManager;

    protected MimeMessageManager mimeMessageManager;

    private Timer timer;

    @Override
    public void init() {
        this.mailSessionManager = ((MailComponent) this.getComponent()).getMailSessionManager();
        this.mimeMessageManager = ((MailComponent) this.getComponent()).getMimeMessageManager();
    }

    /**
     * Send the message to the jbi service defined in the Consume node
     *
     * @param message
     * @throws PEtALSCDKException
     * @throws javax.jbi.messaging.MessagingException
     * @throws IOException
     */
    public void process( Session session, Message message )
    throws javax.jbi.messaging.MessagingException, PEtALSCDKException, IOException, MessagingException {

        this.getLogger().fine("process an incoming email");
        Exchange exchange = this.createConsumeExchange(this.getConsumes(), MEPConstants.IN_ONLY_PATTERN);
        this.setJBIContentAndAttachments( session, message, exchange );
        this.setProperties(message, exchange);
        this.send(exchange);
    }

    /**
     * Sets the body and attachments from the mail {@link Part} to the JBI exchange.
     * <p>
     * If the given message {@link Part} is a {@link Multipart}, all
     * sub part are recursively processed.
     * </p>
     *
     * @param message
     *            the mail {@link Part} to process
     * @param exchange
     *            the resulting jbi exchange
     * @throws MessagingException
     * @throws IOException
     * @throws javax.jbi.messaging.MessagingException
     * @throws PEtALSCDKException
     */
    protected void setJBIContentAndAttachments( Session session, Part message, Exchange exchange )
    throws IOException, MessagingException, javax.jbi.messaging.MessagingException, PEtALSCDKException {

        // Get the message content
        Object content;
        try {
            content = message.getContent();

        } catch( Exception e ) {
            // We can sometimes have troubles with IMAP
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            message.writeTo( bos );
            bos.close();

            SharedByteArrayInputStream bis = new SharedByteArrayInputStream( bos.toByteArray());
            MimeMessage cmsg = new MimeMessage( session, bis );
            bis.close();

            content = cmsg.getContent();
        }

        // Analyze the message content
        if (content instanceof String) {
            if (exchange.getInMessageContentAsSource() == null) {
                String decodedContent = MimeUtility.decodeText( content.toString());

                if( "true".equals(getExtensions().get( MailConstants.IS_XML_CONTENT ))) {
                    this.getLogger().fine("An XML part is set to JBI Exchange content");
                    exchange.setInMessageContent(new StreamSource(new StringReader( decodedContent )));

                } else {
                    this.getLogger().fine("A text part is set to JBI Exchange content (wrap in <body> element)");
                    final Document doc = DocumentBuilders.newDocument();
                    final Element bodyElmt = doc.createElementNS(MailConstants.MAIL_SERVICE_NS, "body");
                    bodyElmt.setTextContent( decodedContent );
                    doc.appendChild(bodyElmt);
                    doc.normalize();
                    exchange.setInMessageContent(doc);
                }
            } else {
                this.getLogger().warning( "A text part is found, but the JBI Exchange content has been already set" );
            }

        } else if (content instanceof Multipart) {
            final Multipart mp = (Multipart) content;
            final int count = mp.getCount();
            for (int i = 0; i < count; i++) {
                this.setJBIContentAndAttachments( session, mp.getBodyPart(i), exchange );
            }

        } else if (content instanceof InputStream) {
            this.getLogger().fine("add an attachment on JBI Exchange");
            exchange.addInMessageAttachment(message.getFileName(), message.getDataHandler());
        }
    }

    /**
     * Set ws addressing values : from, to, action(subject)
     *
     * @param message
     * @param exchange
     * @throws MessagingException
     */
    protected void setProperties(Message message, final Exchange exchange)
            throws MessagingException {

        final StringBuilder s_from = new StringBuilder();
        Address addresses[] = message.getFrom();
        if (addresses != null) {
            for (final Address from : addresses) {
                s_from.append(from.toString()).append(";");
            }
            s_from.deleteCharAt(s_from.length() - 1);
        }

        final StringBuilder s_reply = new StringBuilder();
        addresses = message.getReplyTo();
        if (addresses != null) {
            for (final Address reply : addresses) {
                s_reply.append(reply.toString()).append(";");
            }
            if (s_reply.length() > 0) {
                s_reply.deleteCharAt(s_reply.length() - 1);
            }
        }

        addresses = message.getAllRecipients();
        if (addresses != null) {
            StringBuilder s_to = new StringBuilder();
            for (final Address to : addresses) {
                s_to.append(to.toString()).append(";");
            }
            if (s_to.length() > 0) {
                s_to = s_to.deleteCharAt(s_to.length() - 1);
            }
        }

        final String s_subject = message.getSubject();

        this.getLogger().fine("set property " + MailConstants.MailWSAddressing.FROM_QNAME + " : " + s_from);
        exchange.getInMessage().setProperty(MailConstants.MailWSAddressing.FROM_QNAME.toString(), s_from);
        if (s_reply.length() > 0) {
            this.getLogger().fine("set property " + MailConstants.MailWSAddressing.REPLY_TO_QNAME + " : " + s_reply);
            exchange.getInMessage().setProperty(MailConstants.MailWSAddressing.REPLY_TO_QNAME.toString(), s_reply);
        }
        if (s_subject != null && s_subject.length() > 0) {
            this.getLogger().fine("set property " + MailConstants.MailWSAddressing.ACTION_QNAME + " : " + s_subject);
            exchange.getInMessage().setProperty(MailConstants.MailWSAddressing.ACTION_QNAME.toString(), s_subject);
        }

    }

    @Override
    public void start() throws PEtALSCDKException {
        // Extract session information from SU extensions
        this.extensions = new ConfigurationExtensions(this.getConsumes().getAny());
        try {
            final ConsumeDescriptor consumeDescriptor = MailUtil
                    .buildConsumeDescriptor(this.extensions);

            final ExternalEventPoller listener = new ExternalEventPoller(this.getLogger(), this,
                    getConsumes(), consumeDescriptor, this.mailSessionManager);

            // Create and register the listener
            this.timer = new Timer(this.component.getJbiComponentDescriptor().getComponent().getIdentification()
                    .getName()
                    + " - Incoming mail listener on '"
                    + consumeDescriptor.getUsername()
                    + "@"
                    + consumeDescriptor.getHostname() + ":" + consumeDescriptor.getFolder() + "'");

            // Start listening for new mails
            this.timer.schedule(listener, 0, Integer.parseInt(consumeDescriptor.getPeriod()));

        } catch (MissingElementException e) {
            this.getLogger().warning("Missing element in the jbi.xml : " + e.getElement());
        } catch (javax.jbi.messaging.MessagingException e) {
            throw new PEtALSCDKException(e);
        }

    }

    @Override
    public void stop() throws PEtALSCDKException {
        this.timer.cancel();
    }
}
